Attribute VB_Name = "modNumeralsSorting"
Option Compare Database
Option Explicit

'' **************************************************************************
'' Module:      modNumeralsSorting
'' Author:      Philipp Stiefel (phil@codekabinett.com)
'' Purpose:     Functions for natural numerals sorting in Microsoft Access
'' Version:		1.0.1 	- 2019-10-31 - Bugfix in GetHexString
'' 				1.0		- 2016-08-22 - Initial release
'' Further info: http://codekabinett.com/rdumps.php?Lang=2&targetDoc=logical-numerals-sorting-access-vba
'' **************************************************************************

Private Const ERROR_INSUFFICIENT_BUFFER  As Long = 122
Private Const ERROR_INVALID_FLAGS        As Long = 1004
Private Const ERROR_INVALID_PARAMETER    As Long = 87

Private Const SORT_DIGITSASNUMBERS      As Long = &H8         ''  use digits as numbers sort method
Private Const LCMAP_SORTKEY             As Long = &H400       ''  WC sort key (normalize)

Private Const LOCALE_NAME_USER_DEFAULT      As String = vbNullString
Private Const LOCALE_NAME_INVARIANT         As String = ""
Private Const LOCALE_NAME_SYSTEM_DEFAULT    As String = "!x-sys-default-locale"

Private Declare PtrSafe Function LCMapStringEx Lib "Kernel32.dll" (ByVal lpLocaleName As LongPtr, ByVal dwMapFlags As Long, ByVal lpSrcStr As LongPtr, ByVal cchSrc As Long, _
                                        ByVal lpDestStr As LongPtr, ByVal cchDest As Long, ByVal lpVersionInformation As Long, _
                                        ByVal lpReserved As Long, ByVal sortHandle As Long) As Long

Public Function GetSortKey(ByVal inputString As String) As Byte()

    Dim apiRetVal   As Long
    Dim retVal()    As Byte
    Dim flags       As Long
    
    flags = LCMAP_SORTKEY Or SORT_DIGITSASNUMBERS
    
    If Len(inputString) > 0 Then
        '' call LCMapStringEx with empty buffer to determine the required size of the buffer
        apiRetVal = LCMapStringEx(StrPtr(LOCALE_NAME_SYSTEM_DEFAULT), flags, StrPtr(inputString), Len(inputString), 0, 0, 0, 0, 0)
        
        If apiRetVal > 0 Then
            ReDim retVal(apiRetVal - 1) '' Resize buffer for SortKey
            apiRetVal = LCMapStringEx(StrPtr(LOCALE_NAME_SYSTEM_DEFAULT), flags, StrPtr(inputString), Len(inputString), VarPtr(retVal(0)), UBound(retVal) + 1, 0, 0, 0)
            '' The api writes directly to retVal buffer, we only need to check for an error
            If apiRetVal < 0 Then
                throwError Err.LastDllError, "GetSortKey"
            End If
        Else
            throwError Err.LastDllError, "GetSortKey"
        End If
    Else
        ReDim retVal(0)  '' if input="" return a single zero byte
    End If
    
    GetSortKey = retVal

End Function

Private Sub throwError(ByVal errorCode As Long, ByVal errSource As String)

    Dim errDescription As String

    Select Case errorCode
        Case ERROR_INSUFFICIENT_BUFFER
            errDescription = "Insufficient buffer"
        Case ERROR_INVALID_FLAGS
            errDescription = "Invalid flags"
        Case ERROR_INVALID_PARAMETER
            errDescription = "Invalid parameter"
        Case Else
            errDescription = "unknown error"
    End Select

    Err.Raise vbObjectError + errorCode, errSource, errDescription

End Sub

Private Function GetHexString(buffer() As Byte) As String
    Dim retVal      As String
    Dim i           As Long

    For i = 0 To UBound(buffer)
        retVal = retVal & Right("00" & Hex(buffer(i)), 2)
    Next i

    GetHexString = retVal
End Function

Public Function GetSortKeyHexString(ByVal inputString As String) As String
    GetSortKeyHexString = GetHexString(GetSortKey(inputString))
End Function

Public Function GetSortKeyRawString(ByVal inputString As String) As String
    GetSortKeyRawString = StrConv(GetSortKey(inputString), vbUnicode)
End Function

Public Sub UpdateTableWithSortKey(ByVal tableName As String, ByVal baseColumnName As String, ByVal sortKeyColumnName As String)
    Dim rs As DAO.Recordset
    Dim db As DAO.Database
    
    Set db = CurrentDb
    Set rs = db.OpenRecordset("SELECT " & baseColumnName & ", " & sortKeyColumnName & " FROM " & tableName)
    
    Do Until rs.EOF
        rs.Edit
        rs.Fields(sortKeyColumnName).value = GetSortKey(Nz(rs.Fields(baseColumnName).value, ""))
        rs.Update
        rs.MoveNext
    Loop
    
    rs.Close
    Set rs = Nothing
    Set db = Nothing
    
End Sub

